#include  "axis.h"
#include  <cstdlib>
#include  "lexer.h" 
#include  <algorithm>
#include  "atlas-types.h"

namespace atlas { namespace interpreter {

layer::list layer::lexical_context;

type_expr last_type;
shared_value last_value;

std::shared_ptr<const builtin_value> boolean_negate_builtin;

namespace {

class frame
{
  const id_pat& pattern;
public:
  static shared_context current;

  frame (const id_pat& pattern)
  : pattern(pattern)
  { current = std::make_shared<evaluation_context>(std::move(current)); }
  ~frame() { current = current->tail(); } // don't move here!

  void bind (const shared_value& val)
    { current->reserve(count_identifiers(pattern));
      thread_components(pattern,val,current->back_inserter());
    }
};

class lambda_frame
{
  const id_pat& pattern;
  const shared_context saved;
public:
  lambda_frame (const id_pat& pattern, const shared_context& outer)
  : pattern(pattern), saved(std::move(frame::current))
  { assert(&outer!=&frame::current); // for excluded case use |frame| instead
    try
    { frame::current = std::make_shared<evaluation_context>(outer); }
    catch(...)
    { frame::current = std::move(saved); throw; }
    // restore as destructor would do
  }
  ~lambda_frame() { frame::current = std::move(saved); }

  void bind (const shared_value& val)
    { if (interrupt_flag!=0)
        throw user_interrupt();
      frame::current->reserve(count_identifiers(pattern));
      thread_components(pattern,val,frame::current->back_inserter());
    }
};


shared_context frame::current; 


id_type equals_name()
{ static id_type name=main_hash_table->match_literal("=");
  return name;
}

id_type size_of_name()
{ static id_type name=main_hash_table->match_literal("#");
  return name;
}
id_type print_name()
{ static id_type name=main_hash_table->match_literal("print");
  return name;
}
id_type to_string_name()
{ static id_type name=main_hash_table->match_literal("to_string");
  return name;
}
id_type prints_name()
{ static id_type name=main_hash_table->match_literal("prints");
  return name;
}
id_type error_name()
{ static id_type name=main_hash_table->match_literal("error");
  return name;
}

inline bool is_special_operator(id_type id)
{ return id==size_of_name()
        or id==print_name()
        or id==to_string_name()
        or id==prints_name()
        or id==error_name(); }

bool can_coerce_arg
  (expression e,size_t i,const type_expr& from,type_expr& to)
{ if (to.specialise(from))
    return true; // type matches without coercion
  tuple_expression* tup= dynamic_cast<tuple_expression*>(e);
  if (tup==nullptr or tup->component.size()!=2)
    return false; // we need a pair to insert a coercion
  return coerce(from,to,tup->component[i]);
}

void extend_message
  (error_base& e,const expression_base* function, const value_base* fun,
   const std::string& arg)
{ std::ostringstream o;
  auto bif_p=dynamic_cast<const overloaded_builtin_call*>(function);
  auto clf_p=dynamic_cast<const overloaded_closure_call*>(function);
  auto id_p=dynamic_cast<const identifier*>(function);
  o << "\n(in call of "
    << ( bif_p!=nullptr ? bif_p->print_name
       : clf_p!=nullptr ? clf_p->print_name
       : id_p!=nullptr ? id_p->name() : "anonymous function"
       )
    << ", ";
  auto f=dynamic_cast<const closure_value*>(fun);
  if (f==nullptr)
     o << "built-in";
  else o << "defined " << f->p->loc;
  o << ')';
  if (verbosity>0)
    o << "\n  argument" << (arg[0]=='(' ? "s: " : ": ") << arg;
  e.message.append(o.str());
}

void print_wrapper(expression_base::level l)
{
  *output_stream << *execution_stack.back() << std::endl;
  if (l!=expression_base::single_value) // in |single_value| case we are done
    push_expanded(l,pop_value()); // otherwise remove and possibly expand value
}

std::ostream& to_string_aux(std::ostream& o, expression_base::level l)
{ shared_value v=pop_value();

  const string_value* s=dynamic_cast<const string_value*>(v.get());
  if (s!=nullptr)
    o << s->val; // single string without quotes
  else
  { const tuple_value* t=dynamic_cast<const tuple_value*>(v.get());
    if (t!=nullptr)
    { for (auto it=t->val.begin(); it!=t->val.end(); ++it)
      { s=dynamic_cast<const string_value*>(it->get());
        if (s!=nullptr)
	  o << s->val; // string components without quotes
        else
           o << *it->get(); // treat non-string tuple components as |print|
      }
    }
    else
      o << *v; // output like |print| unless string or tuple
  }
  return o;
}

void to_string_wrapper(expression_base::level l)
{ std::ostringstream o;
  to_string_aux(o,l);
  if (l!=expression_base::no_value)
    push_value(std::make_shared<string_value>(o.str()));
}

void prints_wrapper(expression_base::level l)
{ to_string_aux(*output_stream,l) << std::endl;
  if (l==expression_base::single_value)
    wrap_tuple<0>(); // don't forget to return a value if asked for
}

void error_wrapper(expression_base::level l)
{ std::ostringstream o;
  to_string_aux(o,l);
  throw runtime_error(o.str());
}

void sizeof_wrapper(expression_base::level l)
{ size_t s=get<row_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(s));
}

void suffix_element_wrapper(expression_base::level l)
{ shared_value e=pop_value();
  own_row r=get_own<row_value>();
  if (l!=expression_base::no_value)
  { r->val.push_back(std::move(e));
    push_value(std::move(r));
  }
}

void prefix_element_wrapper(expression_base::level l)
{ own_row r=get_own<row_value>();
  shared_value e=pop_value();
  if (l!=expression_base::no_value)
  { r->val.insert(r->val.begin(),e);
    push_value(std::move(r));
  }
}

void join_rows_wrapper(expression_base::level l)
{ shared_row y=get<row_value>();
  shared_row x=get<row_value>();
  if (l!=expression_base::no_value)
  { own_row result = std::make_shared<row_value>(x->val.size()+y->val.size());
    std::copy(y->val.begin(),y->val.end(), 
      std::copy(x->val.begin(),x->val.end(),result->val.begin()));
  push_value(std::move(result));
  }

}

id_pat copy_id_pat(const id_pat& p)
{
  containers::sl_list<id_pat> rl;
  std::transform(p.sublist.begin(),end(p.sublist)
                , std::back_inserter(rl), copy_id_pat);
  return id_pat (p.name, p.kind, rl.undress());
}

template <  template < unsigned > class  slice  >
expression make_slice(unsigned flags
  ,expression_ptr&& a, expression_ptr&& b, expression_ptr&& c)
{ switch (flags)
  {
  case 0: return new slice<0>(std::move(a),std::move(b),std::move(c));
  case 1: return new slice<1>(std::move(a),std::move(b),std::move(c));
  case 2: return new slice<2>(std::move(a),std::move(b),std::move(c));
  case 3: return new slice<3>(std::move(a),std::move(b),std::move(c));
  case 4: return new slice<4>(std::move(a),std::move(b),std::move(c));
  case 5: return new slice<5>(std::move(a),std::move(b),std::move(c));
  case 6: return new slice<6>(std::move(a),std::move(b),std::move(c));
  case 7: return new slice<7>(std::move(a),std::move(b),std::move(c));
  default: assert(false); return(nullptr);
  }
}

bool was_negated (expr& e)
{ bool negative=false; expr_p p=&e;
  while (p->kind==negation_expr)
  { negative=not negative; p=p->negation_variant; }
  e=std::move(*p); // will do nothing if |p=&e|
  return negative;
}

expression make_while_loop
  (unsigned flags,expression_ptr&& condition,expression_ptr&& body)
{ switch(flags)
  {
  case 0: return new while_expression<0>(std::move(condition),std::move(body));
  case 1: return new while_expression<1>(std::move(condition),std::move(body));
  case 2: return new while_expression<2>(std::move(condition),std::move(body));
  case 3: return new while_expression<3>(std::move(condition),std::move(body));
  default: assert(false); return nullptr;
  }
}

expression make_for_loop
  (unsigned flags, const id_pat& id, expression_ptr&& i, expression_ptr&& b
  ,subscr_base::sub_type t)
{ switch(t)
  {
  case subscr_base::row_entry:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::row_entry>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::row_entry>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::row_entry>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::row_entry>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  case subscr_base::vector_entry:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::vector_entry>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::vector_entry>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::vector_entry>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::vector_entry>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  case subscr_base::ratvec_entry:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::ratvec_entry>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::ratvec_entry>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::ratvec_entry>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::ratvec_entry>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  case subscr_base::string_char:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::string_char>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::string_char>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::string_char>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::string_char>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  case subscr_base::matrix_column:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::matrix_column>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::matrix_column>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::matrix_column>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::matrix_column>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  case subscr_base::mod_poly_term:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::mod_poly_term>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::mod_poly_term>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::mod_poly_term>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::mod_poly_term>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  default: assert(false); return(nullptr);
  }
}

expression make_counted_loop (unsigned flags, id_type id, 
   expression_ptr&& count, expression_ptr&& bound, expression_ptr&& body)
{ switch(flags)
  {
  case 0: return new  counted_for_expression<0>
    (id,std::move(count),std::move(bound),std::move(body));
  case 1: return new  counted_for_expression<1>
    (id,std::move(count),std::move(bound),std::move(body));
  case 2: return new  counted_for_expression<2>
    (id,std::move(count),std::move(bound),std::move(body));
  case 3: return new  counted_for_expression<3>
    (id,std::move(count),std::move(bound),std::move(body));
   case 4: return new  counted_for_expression<4>
    (id,std::move(count),std::move(bound),std::move(body));
  case 6: return new  counted_for_expression<6>
    (id,std::move(count),std::move(bound),std::move(body));
  default: assert(false); return(nullptr);
  }
}

inline bool functype_specialise
  (type_expr& t, const type_expr& from, const type_expr& to)
{ return t.specialise(gen_func_type) 
  and t.func->arg_type.specialise(from) 
  and t.func->result_type.specialise(to);
}

void report_undefined (id_type id,const expr& e,const char* where)
{ std::ostringstream o;
  o << "Undefined identifier '" << main_hash_table->name_of(id)
  << "' in " << where << ' ' << e;
  if (e.loc.file!=Hash_table::empty)
    o << ' ' << e.loc;
  throw program_error (o.str());
}

void report_constant_modified (id_type id,const expr& e,const char* where)
{ std::ostringstream o;
  o << "Name '" << main_hash_table->name_of(id)
    << "' is constant in " << where << ' ' << e;
  if (e.loc.file!=Hash_table::empty)
    o << ' ' << e.loc;
  throw program_error (o.str());
}

}

void reset_evaluator()
{ if (!execution_stack.empty())
  { if (verbosity==0)
      execution_stack.clear();
    else
    {
      std::cerr << "Discarding from execution stack:" << std::endl;
      do
        std::cerr << *pop_value() << std::endl;
      while (!execution_stack.empty());
    }
  }
}

void layer::add(id_type id,type_expr&& t,bool is_const)
{ for (auto it=variable.begin(); it!=variable.end(); ++it)
  // traverse |variable| vector
    if (it->first==id)
      throw program_error 
       (std::string("Multiple binding of '")
                    +main_hash_table->name_of(id)
                    +"' in same scope");
  constness.set_to(variable.size(),is_const);
  variable.emplace_back( id, std::move(t) );
}

const_type_p layer::lookup
  (id_type id, size_t& depth, size_t& offset, bool& is_const)
{ size_t i=0;
  for (auto range=lexical_context.cbegin(); not lexical_context.at_end(range);
       ++range,++i)
    for (auto it=(*range)->cbegin(); it!=(*range)->cend(); ++it)
      if (it->first==id)
      { depth=i;
        offset=it-(*range)->begin();
        is_const = (*range)->is_const(it);
        return &it->second;
      }
  return nullptr;
}

void layer::specialise (size_t depth, size_t offset,const type_expr& t)
{ auto range=lexical_context.cbegin();
  while (depth-->0)
    ++range;
  (**range)[offset].second.specialise(t);
}

expression_ptr convert_expr(const expr& e, type_expr& type)
{
  switch(e.kind)
  {
   
   case integer_denotation:
     { expression_ptr d(new denotation
         (std::make_shared<int_value>(e.int_denotation_variant)));
       return conform_types(int_type,type,std::move(d),e);
     }
   case string_denotation:
     { expression_ptr d(new denotation
         (std::make_shared<string_value>(e.str_denotation_variant)));
       return conform_types(str_type,type,std::move(d),e);
     }
   case boolean_denotation:
     { expression_ptr d(new denotation
           (std::make_shared<bool_value>(e.bool_denotation_variant)));
       return conform_types(bool_type,type,std::move(d),e);
     }
   
   case last_value_computed:
   { std::ostringstream o;
     o << '(' << last_type << ":$)"; 
   return conform_types
       (last_type
       ,type
       ,expression_ptr(new capture_expression(last_value,o.str()))
       ,e);
   }
   
   case die_expr:
   { return expression_ptr(new shell());
   }
   
   case tuple_display:
   { type_expr tup=unknown_tuple(length(e.sublist));
     tuple_expression* tup_exp;
     expression_ptr result(tup_exp=new tuple_expression(0));
     std::vector<expression_ptr>& comp = tup_exp->component;
     comp.reserve(length(e.sublist));
     bool tuple_expected = type.specialise(tup);
     // whether |type| is a tuple of correct size
     wtl_iterator tl_it (tuple_expected ? type.tupple : tup.tupple);
     wel_const_iterator it(e.sublist);
     while (not it.at_end())
       comp.push_back(convert_expr(*it,*tl_it)), ++it,++tl_it;
     if (tuple_expected or coerce(tup,type,result))
         return result;
     throw type_error(e,std::move(tup),std::move(type));
   }
   
   case list_display:
   { std::unique_ptr<list_expression> result (new list_expression(0));
     result->component.reserve(length(e.sublist));
     wel_const_iterator it(e.sublist);
     if (type==void_type)
     { type_expr target; // initially undetermined common component type
       for (; not it.at_end(); ++it)
         result->component.push_back(convert_expr(*it,target));
       return expression_ptr(new voiding(std::move(result)));
       // and forget |target|
     }
     else if (type.specialise(row_of_type))
     { for (; not it.at_end(); ++it)
         result->component.push_back(convert_expr(*it,*type.component_type));
       return std::move(result);
     }
     
     { type_expr comp_type;
       const conversion_record* conv = row_coercion(type,comp_type);
       if (conv==nullptr)
       // no conversion was found; there's nothing left but to report an error
         throw type_error(e,row_of_type.copy(),std::move(type));
     
       for (; not it.at_end(); ++it)
         result->component.push_back(convert_expr(*it,comp_type));
       return expression_ptr(new
         conversion(*conv,expression_ptr(std::move(result))));
     }
   }
   
   case applied_identifier:
   { const id_type id=e.identifier_variant;
     const_type_p id_t; size_t i,j; bool is_const;
     const bool is_local=(id_t=layer::lookup(id,i,j,is_const))!=nullptr;
     if (not is_local and (id_t=global_id_table->type_of(id,is_const))==nullptr)
     {
       std::ostringstream o;
       o << "Undefined identifier '" << main_hash_table->name_of(id) << '\'';
       if (e.loc.file!=Hash_table::empty)
         o << ' ' << e.loc;
       throw program_error (o.str());
     }
   
     expression_ptr id_expr =  is_local
     ? expression_ptr(new local_identifier(id,i,j))
     : expression_ptr(new global_identifier(id));
     if (type.specialise(*id_t)) // then required type admits known identifier type
       { if (type!=*id_t)
         // usage has made type of identifier more specialised
         { if (is_local)
             layer::specialise(i,j,type);
             // then refine its type in the local context
           else
             global_id_table->specialise(id,type); // or in |global_id_table|
         }
         return id_expr;
       }
     else if (coerce(*id_t,type,id_expr))
       return id_expr;
     throw type_error(e,id_t->copy(),std::move(type));
   }
   
   case function_call:
   { if (e.call_variant->fun.kind==applied_identifier)
       
       { const id_type id =e.call_variant->fun.identifier_variant;
         size_t i,j; bool b; // dummies; local binding not used here
         auto local_type_p=layer::lookup(id,i,j,b);
         if (local_type_p==nullptr or local_type_p->kind!=function_type)
        // not calling by local identifier
         { const overload_table::variant_list& variants
             = global_overload_table->variants(id);
           if (variants.size()>0 or is_special_operator(id))
             return resolve_overload(e,type,variants);
         }
       }
     type_expr f_type=gen_func_type.copy(); // start with generic function type
     expression_ptr fun = convert_expr(e.call_variant->fun,f_type);
     expression_ptr arg = convert_expr(e.call_variant->arg,f_type.func->arg_type);
     expression_ptr call (new call_expression(std::move(fun),std::move(arg)));
     return conform_types(f_type.func->result_type,type,std::move(call),e);
   }
   
   case let_expr:
   { const let& lexp=e.let_variant;
     id_pat& pat=lexp->pattern;
     type_expr decl_type=pattern_type(pat);
     expression_ptr arg = convert_expr(lexp->val,decl_type);
   layer new_layer(count_identifiers(pat));
     thread_bindings(pat,decl_type,new_layer,false);
     return expression_ptr(new 
       let_expression(pat,std::move(arg),convert_expr(lexp->body,type)));
   }
   
   case lambda_expr:
   { const lambda& fun=e.lambda_variant;
     const id_pat& pat=fun->pattern;
     type_expr& arg_type=fun->parameter_type;
     if (not arg_type.specialise(pattern_type(pat)))
       throw expr_error(e,"Function argument pattern does not match its type");
   layer new_layer(count_identifiers(pat));
     thread_bindings(pat,arg_type,new_layer,false);
     if (type!=void_type)
     { if (not (type.specialise(gen_func_type)
                and type.func->arg_type.specialise(arg_type)))
       throw type_error(e,
                          type_expr(arg_type.copy(),unknown_type.copy()),
                          std::move(type));
       return expression_ptr(new 
         lambda_expression(pat, convert_expr(fun->body,type.func->result_type)
                          ,std::move(e.loc)));
     }
     else
     { type_expr dummy; // unused result type
       expression_ptr result(new 
         lambda_expression(pat,convert_expr(fun->body,dummy),std::move(e.loc)));
       return expression_ptr(new voiding(std::move(result)));
     }
   }
   
   case seq_expr:
   { const sequence& seq=e.sequence_variant;
     if (seq->forward)
     { expression_ptr first = convert_expr(seq->first,as_lvalue(void_type.copy()));
       expression_ptr last  = convert_expr(seq->last,type);
       return expression_ptr(new
         seq_expression(std::move(first),std::move(last)));
     }
     else
     { expression_ptr first = convert_expr(seq->first,type);
       expression_ptr last  = convert_expr(seq->last,as_lvalue(void_type.copy()));
       return expression_ptr(new
          next_expression(std::move(first),std::move(last)));
     }
   }
   
   case subscription:
   { type_expr array_type, index_type, subscr_type;
       // all initialised to |undetermined_type|
     auto& subsn = *e.subscription_variant;
       // alias |struct| pointed to by the |unique_ptr|
     expression_ptr array = convert_expr(subsn.array,array_type);
     expression_ptr index = convert_expr(subsn.index,index_type);
     expression_ptr subscr;
     
     switch (subscr_base::index_kind(array_type,index_type,subscr_type))
     { case subscr_base::row_entry:
       if (subsn.reversed)
         subscr.reset(new
           row_subscription<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           row_subscription<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::vector_entry:
       if (subsn.reversed)
         subscr.reset(new
           vector_subscription<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           vector_subscription<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::ratvec_entry:
       if (subsn.reversed)
         subscr.reset(new
           ratvec_subscription<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           ratvec_subscription<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::string_char:
       if (subsn.reversed)
         subscr.reset(new
           string_subscription<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           string_subscription<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::matrix_entry:
       if (subsn.reversed)
         subscr.reset(new
           matrix_subscription<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           matrix_subscription<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::matrix_column:
       if (subsn.reversed)
         subscr.reset(new
           matrix_get_column<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           matrix_get_column<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::mod_poly_term:
       if (subsn.reversed)
         throw expr_error(e,"Cannot do reversed subscription of a ParamPol");
       subscr.reset(new module_coefficient(std::move(array),std::move(index)));
     break;
     case subscr_base::not_so:
       std::ostringstream o;
       o << "Cannot subscript value of type " << array_type 
         << " with index of type " << index_type;
       throw expr_error(e,o.str());
     }
     return conform_types(subscr_type,type,std::move(subscr),e);
   }
   
   case slice:
   { type_expr array_type; // initialised to |undetermined_type|
     expression_ptr array = convert_expr(e.slice_variant->array,array_type);
     expression_ptr lower =
       convert_expr(e.slice_variant->lower,as_lvalue(int_type.copy()));
     expression_ptr upper =
       convert_expr(e.slice_variant->upper,as_lvalue(int_type.copy()));
     expression_ptr subscr; const unsigned fl = e.slice_variant->flags.to_ulong();
     switch (subscr_base::slice_kind(array_type))
       { case subscr_base::row_entry: subscr.reset(
        make_slice<row_slice>
           (fl,std::move(array),std::move(lower),std::move(upper)));
       break;
       case subscr_base::vector_entry: subscr.reset(
        make_slice<vector_slice>
           (fl,std::move(array),std::move(lower),std::move(upper)));
       break;
       case subscr_base::ratvec_entry: subscr.reset(
        make_slice<ratvec_slice>
           (fl,std::move(array),std::move(lower),std::move(upper)));
       break;
       case subscr_base::string_char: subscr.reset(
        make_slice<string_slice>
           (fl,std::move(array),std::move(lower),std::move(upper)));
       break;
       case subscr_base::matrix_column: subscr.reset(
        make_slice<matrix_slice>
           (fl,std::move(array),std::move(lower),std::move(upper)));
       break;
       default: std::ostringstream o;
         o << "Cannot slice value of type " << array_type;
         throw expr_error(e,o.str());
       }
   
     return conform_types(array_type,type,std::move(subscr),e);
   }
   
   case negation_expr:
   { type_expr b=bool_type.copy();
      expression_ptr arg = convert_expr(*e.negation_variant,b);
     if (not type.specialise(b)) // |not| preserves the |bool| type
       throw type_error(e,std::move(b),std::move(type));
     return expression_ptr(new overloaded_builtin_call
        (boolean_negate_builtin->val,"{!@bool}",std::move(arg),e.loc));
   }
   
   case conditional_expr:
   { if (was_negated(e.if_variant->condition)) // eliminate negated conditions
       e.if_variant->then_branch.swap(e.if_variant->else_branch);
     expression_ptr c  =
       convert_expr(e.if_variant->condition,as_lvalue(bool_type.copy()));
     type_expr else_type(type.copy());
     // make a copy so as to treat branches similarly
     expression_ptr th = convert_expr(e.if_variant->then_branch,type);
     expression_ptr el = convert_expr(e.if_variant->else_branch,else_type);
     if (type!=else_type) // we had different specialisations in the two branches
     {
       if (type==void_type or else_type==void_type)
         // type was unknown, one branch became void
       {
         if (type==void_type)
           el.reset(new voiding(std::move(el)));
         else
           th.reset(new voiding(std::move(th))),type = std::move(else_type);
       }
       else if (type==unknown_type or else_type==unknown_type)
         // error exit is always OK
         type.specialise(else_type); // but retain the more specific one in |type|
       else
       {
         int cmp = is_close(type,else_type);
         if ((cmp&0x1)!=0) // \.{then} branch may convert to |else_type|
           th =
             convert_expr(e.if_variant->then_branch,type = std::move(else_type));
         else if ((cmp&0x2)!=0) // |else| branch may convert to |type|
           el = convert_expr(e.if_variant->else_branch,type);
         else
         { std::ostringstream o;
           o << "Could not find common type for branches of conditional, "
           "types are " << type << " and " << else_type;
           throw expr_error(e,o.str());
         }
       }
     }
     return expression_ptr(new
       conditional_expression(std::move(c),std::move(th),std::move(el)));
   }
   
   case while_expr:
   { const w_loop& w=e.while_variant;
     if (was_negated(w->condition))
       w->flags.set(0); // this makes an ``until-loop''
     expression_ptr c = convert_expr(w->condition,as_lvalue(bool_type.copy()));
     if (type==void_type)
     { expression_ptr result(make_while_loop  (w->flags.to_ulong(),
          std::move(c),convert_expr(w->body, as_lvalue(void_type.copy()))));
       return expression_ptr(new voiding(std::move(result)));
     }
     else if (type.specialise(row_of_type))
     { expression_ptr b = convert_expr(w->body, *type.component_type);
       return expression_ptr (make_while_loop 
          (w->flags.to_ulong(),std::move(c),std::move(b)));
     }
     else
     
     { type_expr comp_type;
       const conversion_record* conv = row_coercion(type,comp_type);
       if (conv==nullptr)
         throw type_error(e,row_of_type.copy(),std::move(type));
     
       return expression_ptr(new conversion(*conv, expression_ptr(make_while_loop 
            (w->flags.to_ulong(),std::move(c),convert_expr(w->body,comp_type)))));
     }
   }
   
   case for_expr:
   { const f_loop& f=e.for_variant;
     type_expr in_type;
     expression_ptr in_expr = convert_expr(f->in_part,in_type);  // \&{in} part
     subscr_base::sub_type which; // the kind of aggregate iterated over
     layer bind(count_identifiers(f->id));
      // for identifier(s) introduced in this loop
     
     { type_expr comp_type; const type_expr* tp;
       which = subscr_base::index_kind(in_type,*(tp=&int_type),comp_type);
       if (which==subscr_base::not_so)
         which = subscr_base::index_kind(in_type,*(tp=&param_type),comp_type);
       if (which==subscr_base::not_so)
       { std::ostringstream o;
         o << "Cannot iterate over value of type " << in_type;
         throw expr_error(e,o.str());
       }
       type_expr pt = pattern_type(f->id);
       type_list it_comps;
       it_comps.push_front(std::move(comp_type));
       it_comps.push_front(type_expr(tp->copy()));
       type_expr it_type(std::move(it_comps));
       if (not pt.specialise(it_type))
         throw expr_error(e,"Improper structure of loop variable pattern");
       thread_bindings(f->id,it_type,bind,true); // force all identifiers constant
     }
     type_expr body_type;
     type_expr *btp=&body_type; // point to place to record body type
     const conversion_record* conv=nullptr;
     if (type==void_type)
       btp=&type; // we can reuse this type; no risk of specialisation
     else if (type.specialise(row_of_type))
       btp=type.component_type;
     else if ((conv=row_coercion(type,body_type))==nullptr)
       throw type_error(e,row_of_type.copy(),std::move(type));
     expression_ptr body(convert_expr (f->body,*btp));
   expression_ptr loop(make_for_loop
       (f->flags.to_ulong(),f->id,std::move(in_expr),std::move(body),which));
   return type==void_type ? expression_ptr(new voiding(std::move(loop)))
     :  conv!=nullptr ? expression_ptr(new conversion(*conv,std::move(loop)))
     :  std::move(loop) ;
   }
   
   case cfor_expr:
   { const c_loop& c=e.cfor_variant;
     expression_ptr count_expr = convert_expr(c->count,as_lvalue(int_type.copy()));
     static const shared_value zero = std::make_shared<int_value>(0);
       // avoid repeated allocation
     expression_ptr bound_expr = is_empty(c->bound) 
       ? expression_ptr(new denotation(zero))
       : convert_expr(c->bound,as_lvalue(int_type.copy())) ;
   
     type_expr body_type;
     type_expr *btp=&body_type; // point to place to record body type
     const conversion_record* conv=nullptr;
     if (type==void_type)
       btp=&type; // we can reuse this type; no risk of specialisation
     else if (type.specialise(row_of_type))
       btp=type.component_type;
     else if ((conv=row_coercion(type,body_type))==nullptr)
       throw type_error(e,row_of_type.copy(),std::move(type));
   
     if (c->flags[2]) // case of no loop variable; avoid introducing empty |layer|
     { expression_ptr body(convert_expr (c->body,*btp));
     expression_ptr loop(make_counted_loop(c->flags.to_ulong(), 
         c->id,std::move(count_expr),std::move(bound_expr),std::move(body)));
     return type==void_type ? expression_ptr(new voiding(std::move(loop))) : 
            conv!=nullptr ? expression_ptr(new conversion(*conv,std::move(loop)))
                          :  std::move(loop);
     }
     else // case of a loop variable
     { layer bind(1);
       bind.add(c->id,int_type.copy(),true); // add |id| as constant
       expression_ptr body(convert_expr (c->body,*btp));
     expression_ptr loop(make_counted_loop(c->flags.to_ulong(), 
         c->id,std::move(count_expr),std::move(bound_expr),std::move(body)));
       return type==void_type ? expression_ptr(new voiding(std::move(loop))) : 
            conv!=nullptr ? expression_ptr(new conversion(*conv,std::move(loop)))
                          :  std::move(loop);
     }
   }
   
   case cast_expr:
   { const cast& c=e.cast_variant;
     expression_ptr p = convert_expr(c->exp,c->type);
     return conform_types(c->type,type,std::move(p),e);
   }
   
   case op_cast_expr:
   { const op_cast& c=e.op_cast_variant;
     const overload_table::variant_list& variants =
      global_overload_table->variants(c->oper);
     type_expr& ctype=c->type;
     if (is_special_operator(c->oper))
       
       { if (c->oper==print_name())
         { if (functype_specialise(type,ctype,ctype))
           return expression_ptr(new  denotation
             (std::make_shared<builtin_value>(print_wrapper,"print")));
         }
         else if (c->oper==prints_name())
         { if (functype_specialise(type,ctype,void_type))
           return expression_ptr(new  denotation
             (std::make_shared<builtin_value>(prints_wrapper,"prints")));
         }
         else if (c->oper==to_string_name())
         { if (functype_specialise(type,ctype,str_type))
           return expression_ptr(new  denotation
             (std::make_shared<builtin_value>(prints_wrapper,"to_string")));
         }
         else if (c->oper==error_name())
         { if (functype_specialise(type,ctype,unknown_type))
           return expression_ptr(new  denotation
             (std::make_shared<builtin_value>(prints_wrapper,"error")));
         }
         else if (c->oper==size_of_name())
           
           { if (ctype.kind==row_type)
             { if (functype_specialise(type,ctype,int_type))
             return expression_ptr(new  denotation
                 (std::make_shared<builtin_value>(sizeof_wrapper,"#@[T]")));
               throw type_error(e,ctype.copy(),std::move(type));
             }
             else if (ctype.specialise(pair_type))
             {
               type_expr& arg_tp0 = ctype.tupple->contents;
               type_expr& arg_tp1 = ctype.tupple->next->contents;
               if (arg_tp0.kind==row_type)
               { if (arg_tp0==arg_tp1)
                 { if (functype_specialise(type,ctype,arg_tp0))
                   return expression_ptr(new  denotation
                     (std::make_shared<builtin_value>(join_rows_wrapper
                                                     ,"#@([T],[T]->[T])")));
                   throw type_error(e,ctype.copy(),std::move(type));
                 }
                 else if (*arg_tp0.component_type==arg_tp1)
                 { if (functype_specialise(type,ctype,arg_tp0))
                   return expression_ptr(new  denotation
                       (std::make_shared<builtin_value>(suffix_element_wrapper
                                                         ,"#@([T],T->[T])")));
                   throw type_error(e,ctype.copy(),std::move(type));
                 }
               }
               if (arg_tp1.kind==row_type and *arg_tp1.component_type==arg_tp0)
               { if (functype_specialise(type,ctype,arg_tp1))
                 return expression_ptr(new  denotation
                   (std::make_shared<builtin_value>(prefix_element_wrapper
                                                     ,"#@(T,[T]->[T])")));
                 throw type_error(e,ctype.copy(),std::move(type));
               }
             }
           }
       }
     size_t i;  std::ostringstream o;
     for (i=0; i<variants.size(); ++i)
       if (variants[i].type().arg_type==ctype)
         break;
     if (i==variants.size()) // nothing was found
     {
       o << "Cannot resolve " << main_hash_table->name_of(c->oper) 
          << " at argument type " << ctype;
     throw program_error(o.str());
     }
     o << main_hash_table->name_of(c->oper) << '@' << ctype;
     expression_ptr p(new capture_expression(variants[i].val,o.str()));
     const type_expr& res_t = variants[i].type().result_type;
     if (functype_specialise(type,ctype,res_t))
       return p;
     else if (type==void_type)
       return expression_ptr(new voiding(std::move(p)));
     else throw
         type_error(e,type_expr(ctype.copy(),res_t.copy()),std::move(type));
   }
   break;
   
   case ass_stat:
   {
     id_type lhs=e.assign_variant->lhs;
     const_type_p id_t; size_t i,j; bool is_const;
     const bool is_local = (id_t=layer::lookup(lhs,i,j,is_const))!=nullptr;
     if (not is_local and (id_t=global_id_table->type_of(lhs,is_const))==nullptr)
       report_undefined(lhs,e,"assignment");
   
     if (is_const)
       report_constant_modified(lhs,e,"assignment");
   
   
     type_expr rhs_type = id_t->copy(); // provide a modifiable copy
     expression_ptr r(convert_expr(e.assign_variant->rhs,rhs_type));
     if (rhs_type!=*id_t)
       // assignment will specialise identifier, record to which type it does
     { if (is_local)
         layer::specialise(i,j,rhs_type);
         else
         global_id_table->specialise(lhs,rhs_type);
     }
   
     expression_ptr assign = is_local
     ? expression_ptr(new local_assignment(lhs,i,j,std::move(r)))
   : expression_ptr(new global_assignment(lhs,std::move(r)));
     return conform_types(rhs_type,type,std::move(assign),e);
   }
   
   case comp_ass_stat:
   { id_type aggr=e.comp_assign_variant->aggr;
     const expr& index=e.comp_assign_variant->index;
     const expr& rhs=e.comp_assign_variant->rhs;
   const_type_p aggr_t; size_t d,o; bool is_const;
     bool is_local = (aggr_t=layer::lookup(aggr,d,o,is_const))!=nullptr;
     if (not is_local and (aggr_t=global_id_table->type_of(aggr,is_const))==nullptr)
       report_undefined(aggr,e,"component assignment");
   
     if (is_const)
       report_constant_modified(aggr,e,"component assignment");
   
   
     type_expr ind_t;
     expression_ptr i = convert_expr(index,ind_t);
   type_expr comp_t;
     subscr_base::sub_type kind=subscr_base::index_kind(*aggr_t,ind_t,comp_t);
     if (not subscr_base::assignable(kind))
     { std::ostringstream o;
       o << "Cannot subscript value of type " << *aggr_t 
         << " with index of type " << ind_t << " in assignment";
       throw expr_error(e,o.str());
     }
     expression_ptr r = convert_expr(rhs,comp_t);
     if (aggr_t->kind==row_type)
       aggr_t->component_type->specialise(comp_t); // record type
     expression_ptr p;
     if (is_local)
       if (e.comp_assign_variant->reversed)
         p.reset(new local_component_assignment<true>
           (aggr,std::move(i),d,o,std::move(r),kind));
       else
         p.reset(new local_component_assignment<false>
           (aggr,std::move(i),d,o,std::move(r),kind));
     else
       if (e.comp_assign_variant->reversed)
         p.reset(new global_component_assignment<true>
           (aggr,std::move(i),std::move(r),kind));
       else
         p.reset(new global_component_assignment<false>
           (aggr,std::move(i),std::move(r),kind));
     return conform_types(comp_t,type,std::move(p),e);
   }
   case no_expr: assert(false);
  }
  return expression_ptr(nullptr); // keep compiler happy
}

void shell::evaluate (level l) const
{ throw runtime_error("I die"); // |shell| explodes
}

void tuple_expression::print(std::ostream& out) const
{ out << '(';
  for (auto it=component.begin(); it!=component.end(); ++it)
    out << (it==component.begin() ? "" : ",") << **it;
  out << ')';
}

void tuple_expression::evaluate(level l) const
{ if (l==no_value)
    for (auto it=component.begin(); it!=component.end(); ++it)
      (*it)->void_eval();
  else
  { for (auto it=component.begin(); it!=component.end(); ++it)
      (*it)->eval();
    if (l==single_value)
      wrap_tuple(component.size());
  }
}

void list_expression::print(std::ostream& out) const
{ out << '[';
  for (auto it=component.begin(); it!=component.end(); ++it)
    out << (it==component.begin() ? "" : ",") << **it;
  out << ']';
}

void list_expression::evaluate(level l) const
{ if (l==no_value)
    for (auto it=component.begin(); it!=component.end(); ++it)
      (*it)->void_eval();
  else
  { own_row result= std::make_shared<row_value>(0);
    result->val.reserve(component.size());
    for (auto it=component.begin(); it!=component.end(); ++it)
  {
      (*it)->eval();
      result->val.push_back(pop_value());
    }
    push_value(std::move(result)); // result will be shared from here on
  }
}

const char* identifier::name() const
{ return main_hash_table->name_of(code); }
void identifier::print(std::ostream& out) const
{ out<< name(); }

global_identifier::global_identifier(id_type id)
: identifier(id), address(global_id_table->address_of(id))
{}

void global_identifier::evaluate(level l) const
{ if (address->get()==nullptr)
  { std::ostringstream o;
    o << "Taking value of uninitialized variable '" << name() << '\'';
    throw runtime_error(o.str());
  }
  push_expanded(l,*address);
}

void local_identifier::evaluate(level l) const
{ push_expanded(l,frame::current->elem(depth,offset)); }

expression_ptr resolve_overload
  (const expr& e,
   type_expr& type,
   const overload_table::variant_list& variants)
{ const expr& args = e.call_variant->arg;
  type_expr a_priori_type;
  expression_ptr arg = convert_expr(args,a_priori_type);
    // get \foreign{a priori} type or argument
  id_type id =  e.call_variant->fun.identifier_variant;
  
  { if (id==size_of_name())
    { if (a_priori_type.kind==row_type)
      { expression_ptr call(new 
          overloaded_builtin_call(sizeof_wrapper,"#",std::move(arg),e.loc));
        return conform_types(int_type,type,std::move(call),e);
      }
      else if (a_priori_type.kind!=undetermined_type and
               a_priori_type.specialise(pair_type))
      
      {
        type_expr& arg_tp0 = a_priori_type.tupple->contents;
        type_expr& arg_tp1 = a_priori_type.tupple->next->contents;
        if (arg_tp0.kind==row_type)
        { if (can_coerce_arg(arg.get(),1,arg_tp1,*arg_tp0.component_type)) // suffix
          { expression_ptr call(new  overloaded_builtin_call
              (suffix_element_wrapper,"#",std::move(arg),e.loc));
            return conform_types(arg_tp0,type,std::move(call),e);
          }
          if (arg_tp0==arg_tp1) // join
          { expression_ptr call(new  overloaded_builtin_call
              (join_rows_wrapper,"#",std::move(arg),e.loc));
            return conform_types(arg_tp0,type,std::move(call),e);
          }
        }
        if (arg_tp1.kind==row_type and 
               can_coerce_arg(arg.get(),0,arg_tp0,*arg_tp1.component_type))
                // prefix
        { expression_ptr call(new  overloaded_builtin_call
            (prefix_element_wrapper,"#",std::move(arg),e.loc));
          return conform_types(arg_tp1,type,std::move(call),e);
        }
      }
    }
    else if (id==print_name()) // this one always matches
    { expression_ptr call(new
        generic_builtin_call(print_wrapper,"print",std::move(arg),e.loc));
      return conform_types(a_priori_type,type,std::move(call),e);
   }
    else if(id==to_string_name()) // this always matches as well
    { expression_ptr call(new
        generic_builtin_call(to_string_wrapper,"to_string",std::move(arg),e.loc));
      if (type.specialise(str_type))
        return call;
      throw type_error(e,str_type.copy(),std::move(type));
    }
    else if(id==prints_name()) // this always matches as well
    { expression_ptr call(new
        generic_builtin_call(prints_wrapper,"prints",std::move(arg),e.loc));
      if (type.specialise(void_type))
        return call;
      throw type_error(e,void_type.copy(),std::move(type));
    }
    else if(id==error_name()) // this always matches as well
    { return expression_ptr(new
        generic_builtin_call(error_wrapper,"error",std::move(arg),e.loc));
    }
  }
  for (auto it=variants.begin(); it!=variants.end(); ++it)
  { const overload_data& v=*it;
    bool match = a_priori_type==v.type().arg_type;
    if (not match and v.type().arg_type!=void_type)
    {
      if ((is_close(a_priori_type,v.type().arg_type)&0x1)!=0)
      // could first convert to second?
      { type_expr arg_t = v.type().arg_type.copy(); // need a modifiable value
        arg=convert_expr(args,arg_t); // redo conversion
        match=true;
      }
    }
    if (match)
      
      { expression_ptr call;
        const builtin_value* f = dynamic_cast<const builtin_value*>(v.val.get());
        if (f!=nullptr)
          call = expression_ptr (new  overloaded_builtin_call
            (f->val,f->print_name.c_str(),std::move(arg),e.loc));
        else
          
          { shared_closure fun = std::dynamic_pointer_cast<const closure_value>(v.val);
            if (fun==nullptr)
              throw logic_error("Overloaded value is not a function");
            std::ostringstream name;
            name << main_hash_table->name_of(id) << '@' << v.type().arg_type;
            call =
              expression_ptr(new
                overloaded_closure_call(fun,name.str(),std::move(arg),e.loc));
          }
        if (type==void_type and
            id==equals_name() and
            v.type().result_type!=void_type)
          throw type_error(e,v.type().result_type.copy(),std::move(type));
        return conform_types(v.type().result_type,type,std::move(call),e);
      }
  }

  
  if (variants.size()==1)
    throw type_error(args,a_priori_type.copy(),variants[0].type().arg_type.copy());
  else
  { std::ostringstream o;
    o << "Failed to match '"
      << main_hash_table->name_of(id) 
      << "' with argument type "
      << a_priori_type;
    throw expr_error(e,o.str());
  }
}

void call_expression::print(std::ostream& out) const
{ if (dynamic_cast<identifier*>(function.get())!=nullptr)
    out << *function;
  else out << '(' << *function << ')';
  if (dynamic_cast<tuple_expression*>(argument.get())!=nullptr)
    out << *argument;
  else out << '(' << *argument << ')';
}

void overloaded_builtin_call::print(std::ostream& out) const
{ out << print_name;
  if (dynamic_cast<tuple_expression*>(argument.get())!=nullptr)
    out << *argument;
  else out << '(' << *argument << ')';
}

void call_expression::evaluate(level l) const
{ function->eval();  shared_value fun=pop_value();
size_t sp = execution_stack.size();
  std::string arg_string;
  const builtin_value* f=dynamic_cast<const builtin_value*>(fun.get());
  argument->evaluate(f==nullptr ? single_value : multi_value);
  if (verbosity>0)
  { std::ostringstream o;
    if (f==nullptr or sp+1==execution_stack.size()) // get single argument
      o << *execution_stack.back();
    else // built-in with multiple arguments, gather them
    { o << '(';
      while(sp<execution_stack.size())
    { o << *execution_stack[sp++];
          o << (sp<execution_stack.size() ? ',' : ')');
      }
    }
    arg_string = o.str();
  }

  try
  { if (f==nullptr)
      
      { const closure_value* f=force<closure_value>(fun.get());
      
        lambda_frame fr(f->p->param,f->context);
          // save context, create new one for |f|
        fr.bind(pop_value()); // decompose arguments(s) and bind values in |fr|
        f->body.evaluate(l); // call, passing evaluation level |l| to function body
      } 
    else // built-in functions
      (*f->val)(l); // call the wrapper function, handling |l| appropriately
  }
  
  catch (error_base& e)
  { extend_message(e,function.get(),fun.get(),arg_string);
    throw;
  }
  catch (const std::exception& e)
  { runtime_error new_error(e.what());
    extend_message(new_error,function.get(),fun.get(),arg_string);
    throw new_error;
  }
}

void overloaded_builtin_call::evaluate(level l) const
{ size_t sp = execution_stack.size();
  std::string arg_string;
  argument->multi_eval();
  if (verbosity>0) // then record argument(s) as string
  { std::ostringstream o;
    if (sp+1==execution_stack.size())
      o << *execution_stack.back();
    else
    { o << '(';
      while(sp<execution_stack.size())
    { o << *execution_stack[sp++];
          o << (sp<execution_stack.size() ? ',' : ')');
      }
    }
    arg_string = o.str();
  }

  
  try
  { (*f)(l); }
  catch (error_base& e)
  { extend_message(e,this,nullptr,arg_string);
    throw;
  }
  catch (const std::exception& e)
  { runtime_error new_error(e.what());
    extend_message(new_error,this,nullptr,arg_string);
    throw new_error;
  }
}

void generic_builtin_call::evaluate(level l) const
{ std::string arg_string;
  argument->eval();
  if (verbosity>0) // then record argument(s) as string
  { std::ostringstream o;
    o << *execution_stack.back();
    arg_string = o.str();
  }

  
  try
  { (*f)(l); }
  catch (error_base& e)
  { extend_message(e,this,nullptr,arg_string);
    throw;
  }
  catch (const std::exception& e)
  { runtime_error new_error(e.what());
    extend_message(new_error,this,nullptr,arg_string);
    throw new_error;
  }
}

inline
let_expression::let_expression 
  (const id_pat& v, expression_ptr&& ini, expression_ptr&& b)
: variable(copy_id_pat(v))
, initialiser(ini.release())
, body(b.release())
{}

void let_expression::print(std::ostream& out) const
{ out << " let " << variable << "=" << *initialiser << " in " << *body;
}

type_list pattern_list_types(const patlist& p)
{ dressed_type_list result;
  for (auto it = p.begin(); not p.at_end(it); ++it)
    result.push_back(pattern_type(*it));
  return result.undress();
}

type_expr pattern_type(const id_pat& pat)
{ return (pat.kind&0x2)==0
  ? unknown_type.copy()
  : type_expr(pattern_list_types(pat.sublist));
}

size_t count_identifiers(const id_pat& pat)
{ size_t result= pat.kind & 0x1; // 1 if |pat.name| is defined, 0 otherwise
  if ((pat.kind & 0x2)!=0) // then a list of subpatterns is present
    for (auto it=pat.sublist.begin(); not pat.sublist.at_end(it); ++it)
      result+=count_identifiers(*it);
  return result;
}

void list_identifiers(const id_pat& pat, std::vector<id_type>& d)
{ if ((pat.kind & 0x1)!=0)
    d.push_back(pat.name);
  if ((pat.kind & 0x2)!=0) // then a list of subpatterns is present
    for (auto it=pat.sublist.begin(); not pat.sublist.at_end(it); ++it)
      list_identifiers(*it,d);
}

void thread_bindings
(const id_pat& pat,const type_expr& type, layer& dst, bool is_const)
{ if ((pat.kind & 0x1)!=0)
    dst.add(pat.name,type.copy(),is_const or (pat.kind & 0x4)!=0);
  if ((pat.kind & 0x2)!=0)
  { assert(type.kind==tuple_type);
    wtl_const_iterator t_it(type.tupple);
    for (auto p_it=pat.sublist.begin(); not pat.sublist.at_end(p_it);
         ++p_it,++t_it)
      thread_bindings(*p_it,*t_it,dst,is_const);
  }
}

void thread_components
  (const id_pat& pat,const shared_value& val
  , std::back_insert_iterator<std::vector<shared_value> > dst)
{ if ((pat.kind & 0x1)!=0)
     *dst++ = val; // copy |shared_value| pointer |val|, creating sharing

  if ((pat.kind & 0x2)!=0)
  { const tuple_value* t=force<tuple_value>(val.get());
    assert(t->val.size()==length(pat.sublist));
    auto src = t->val.begin();
    for (auto it=pat.sublist.begin(); not pat.sublist.at_end(it); ++it,++src)
      thread_components(*it,*src,dst);
  }
}

void let_expression::evaluate(level l) const
{ initialiser->eval(); // evaluate on stack as single value

  frame fr(variable); // save context, create new one for |f|
  fr.bind(pop_value()); // decompose arguments(s) and bind values in |fr|
  body->evaluate(l); // call, passing evaluation level |l| to function body
} 

inline
lambda_expression::lambda_expression 
  (const id_pat& p, expression_ptr&& b, const source_location& loc)
: p(std::make_shared<lambda_struct>(copy_id_pat(p),std::move(b),loc))
{}

std::ostream& operator<<(std::ostream& out, const lambda_struct& l)
{ if ((l.param.kind&0x1)!=0)
    if ((l.param.kind&0x4)!=0)
      out << "(!" << l.param << ')';
    else out << '(' << l.param << ')';
  else if ((l.param.kind&0x2)!=0 and not l.param.sublist.empty())
    out << l.param;
  else out << '@';
  return out << ": " << *l.body;
}
void lambda_expression::print(std::ostream& out) const { out << *p; }

void closure_value::print(std::ostream& out) const
{ out << "Function defined " << p->loc << std::endl << *p;
}

void lambda_expression::evaluate(level l) const
{if (l!=no_value)
     push_value(std::make_shared<closure_value>(frame::current,p));
}

void overloaded_closure_call::print(std::ostream& out) const
{ out << print_name;
  if (dynamic_cast<tuple_expression*>(argument.get())!=nullptr)
     out << *argument;
  else out << '(' << *argument << ')';
}

void overloaded_closure_call::evaluate(level l) const
{ argument->eval();
  std::string arg_string;
  try
  { lambda_frame fr(fun->p->param,fun->context);
    if (verbosity>0) // then record argument(s) as string
    { std::ostringstream o;
      o << *execution_stack.back();
      arg_string = o.str();
    }

    // now save context, create new one for |fun|
    fr.bind(pop_value()); // decompose arguments(s) and bind values in |fr|
    fun->body.evaluate(l);
    // call, passing evaluation level |l| to function body
  }
  catch (error_base& e)
  { extend_message(e,this,fun.get(),arg_string);
    throw;
  }
  catch (const std::exception& e)
  { runtime_error new_error(e.what());
    extend_message(new_error,this,fun.get(),arg_string);
    throw new_error;
  }
}

void seq_expression::print(std::ostream& out) const
{ out << *first << ';' << *last; }
void next_expression::print(std::ostream& out) const
{ out << *first << " next " << *last; }

void seq_expression::evaluate(level l) const
{ first->void_eval(); last->evaluate(l); }

void next_expression::evaluate(level l) const
{ first->evaluate(l); last->void_eval(); }

void subscr_base::print(std::ostream& out,bool reversed) const
{ out << *array << (reversed ? "~[" : "[") << *index << ']';
}

template <bool reversed>
void matrix_subscription<reversed>::print(std::ostream& out) const
{ out  << *array << (reversed ? "~[" : "[");
  tuple_expression* p=dynamic_cast<tuple_expression*>(index.get());
  if (p==nullptr) out << *index << ']';
  else
    out << *p->component[0] << ',' << *p->component[1] << ']';
}

void slice_base::print(std::ostream& out, unsigned flags) const
{  out << *array << ((flags&0x1)==0 ? "[" : "~[")
     << *lower << ((flags&0x2)==0 ? ":" : "~:")
     << *upper << ((flags&0x4)==0 ? "]" : "~]");
}

subscr_base::sub_type subscr_base::index_kind
  (const type_expr& aggr,
   const type_expr& index,
         type_expr& subscr)
{ if (aggr.kind==primitive_type)
    switch (aggr.prim)
    {  default:
    break; case vector_type:
      if (index==int_type and subscr.specialise(int_type))
        return vector_entry; 
    break; case rational_vector_type:
      if (index==int_type and subscr.specialise(rat_type))
        return ratvec_entry; 
    break; case string_type:
      if (index==int_type and subscr.specialise(str_type))
        return string_char; 
    break; case matrix_type:
      if (index==int_int_type and subscr.specialise(int_type))
        return matrix_entry;
      else if (index==int_type and subscr.specialise(vec_type))
        return matrix_column;
    break; case virtual_module_type:
      if (index==param_type and subscr.specialise(split_type))
        return mod_poly_term;
    }
  else if (aggr.kind==row_type and index==int_type and
           subscr.specialise(*aggr.component_type))
         return row_entry;
  return not_so;
}

subscr_base::sub_type subscr_base::slice_kind (const type_expr& aggr)
{ if (aggr.kind==primitive_type)
    switch (aggr.prim)
    {
    case vector_type: return vector_entry;
    case rational_vector_type: return ratvec_entry;
    case string_type: return string_char;
    case matrix_type: return matrix_column;
    default: return not_so;
    }
  else if (aggr.kind==row_type)
    return row_entry;
  else return not_so;
}

bool subscr_base::assignable(subscr_base::sub_type t)
{ switch (t)
  { case ratvec_entry: case string_char: case mod_poly_term:
    case not_so: return false;
    default: return true;
  }
}

inline std::string range_mess(int i,size_t n,const expression_base* e)
{ std::ostringstream o;
  e->print(o << "index " << i << " out of range (0<= . <" << n
             << ") in subscription ");
  return o.str();
}

template <bool reversed>
void row_subscription<reversed>::evaluate(level l) const
{ int i=(index->eval(),get<int_value>()->val);
  shared_row r=(array->eval(),get<row_value>());
  size_t n = r->val.size();
  if (reversed)
    i=n-1-i;
  if (static_cast<unsigned int>(i)>=n)
    throw runtime_error(range_mess(i,n,this));
  push_expanded(l,r->val[i]);
}

template <bool reversed>
void vector_subscription<reversed>::evaluate(level l) const
{ int i=(index->eval(),get<int_value>()->val);
  shared_vector v=(array->eval(),get<vector_value>());
  size_t n = v->val.size();
  if (reversed)
    i=n-1-i;
  if (static_cast<unsigned int>(i)>=n)
    throw runtime_error(range_mess(i,n,this));
  if (l!=no_value)
    push_value(std::make_shared<int_value>(v->val[i]));
}

template <bool reversed>
void ratvec_subscription<reversed>::evaluate(level l) const
{ int i=(index->eval(),get<int_value>()->val);
  shared_rational_vector v=(array->eval(),get<rational_vector_value>());
  size_t n = v->val.size();
  if (reversed)
    i=n-1-i;
  if (static_cast<unsigned int>(i)>=n)
    throw runtime_error(range_mess(i,n,this));
  if (l!=no_value)
    push_value(std::make_shared<rat_value>(Rational 
       (v->val.numerator()[i],v->val.denominator())));
}

template <bool reversed>
void string_subscription<reversed>::evaluate(level l) const
{ int i=(index->eval(),get<int_value>()->val);
  shared_string s=(array->eval(),get<string_value>());
  size_t n = s->val.size();
  if (reversed)
    i=n-1-i;
  if (static_cast<unsigned int>(i)>=n)
    throw runtime_error(range_mess(i,n,this));
  if (l!=no_value)
    push_value(std::make_shared<string_value>(s->val.substr(i,1)));
}

template <bool reversed>
void matrix_subscription<reversed>::evaluate(level l) const
{ index->multi_eval(); 
  int j=get<int_value>()->val;
  int i=get<int_value>()->val;
  shared_matrix m=(array->eval(),get<matrix_value>());
  size_t r = m->val.numRows();
  size_t c = m->val.numColumns();
  if (reversed)
  {  i=r-1-i; j=c-1-j; }
  if (static_cast<unsigned int>(i)>=r)
    throw runtime_error
     ("initial "+range_mess(i,r,this));
  if (static_cast<unsigned int>(j)>=c)
    throw runtime_error
     ("final "+range_mess(j,c,this));
  if (l!=no_value)
    push_value(std::make_shared<int_value>(m->val(i,j)));
}

template <bool reversed>
void matrix_get_column<reversed>::evaluate(level l) const
{ int j=(index->eval(),get<int_value>()->val);
  shared_matrix m=(array->eval(),get<matrix_value>());
  size_t c = m->val.numColumns();
  if (reversed)
    j=c-1-j;
  if (static_cast<unsigned int>(j)>=c)
    throw runtime_error(range_mess(j,c,this));
  if (l!=no_value)
    push_value(std::make_shared<vector_value>(m->val.column(j)));
}

void slice_range_error
  (int lwb, int upb,int n,unsigned flags,const expression_base* e)
{ std::ostringstream o;
  if (upb>n)
    if (lwb<0)
      o << "both bounds " << lwb << ':' << upb;
    else
      o << "upper bound " << upb;
  else
    o << "lower bound " << lwb;
  o << " out of range (should be ";
  if (lwb<0)
    o << ">=0" << (upb>n ? " respectively " : ")");
  if (upb>n)
    o << "<=" << n << ')';
  e->print(o << " in slice ");
  throw runtime_error(o.str());
}

template <unsigned flags>
void row_slice<flags>::evaluate(level l) const
{ int upb=(upper->eval(),get<int_value>()->val);
  int lwb=(lower->eval(),get<int_value>()->val);
  shared_row arr=(array->eval(),get<row_value>());
  const auto& r = arr->val;
  int n = r.size();
  if ((flags&0x2)!=0)
    lwb = n - lwb;
  if ((flags&0x4)!=0)
    upb = n - upb;
  if (lwb<0 or upb>n)
    slice_range_error(lwb,upb,n,flags,this);
  if (lwb>=upb)
  { push_value(std::make_shared<row_value>(0)); return; }
  push_value( (flags&0x1)==0
  ? std::make_shared<row_value>(r.begin()+lwb,r.begin()+upb)
  : std::make_shared<row_value>(r.rbegin()+lwb,r.rbegin()+upb)
    );
}

template <unsigned flags>
void vector_slice<flags>::evaluate(level l) const
{ int upb=(upper->eval(),get<int_value>()->val);
  int lwb=(lower->eval(),get<int_value>()->val);
  shared_vector arr=(array->eval(),get<vector_value>());
  const auto& r = arr->val;
  int n = r.size();
  if ((flags&0x2)!=0)
    lwb = n - lwb;
  if ((flags&0x4)!=0)
    upb = n - upb;
  if (lwb<0 or upb>n)
    slice_range_error(lwb,upb,n,flags,this);
  if (lwb>=upb)
  { push_value(std::make_shared<vector_value>(int_Vector(0))); return; }
  push_value( (flags&0x1)==0
  ? std::make_shared<vector_value>(r.begin()+lwb,r.begin()+upb)
  : std::make_shared<vector_value>(r.rbegin()+lwb,r.rbegin()+upb)
    );
}

template <unsigned flags>
void ratvec_slice<flags>::evaluate(level l) const
{ int upb=(upper->eval(),get<int_value>()->val);
  int lwb=(lower->eval(),get<int_value>()->val);
  shared_rational_vector arr=(array->eval(),get<rational_vector_value>());
  const auto& r = arr->val.numerator();
  int n = r.size();
  if ((flags&0x2)!=0)
    lwb = n - lwb;
  if ((flags&0x4)!=0)
    upb = n - upb;
  if (lwb<0 or upb>n)
    slice_range_error(lwb,upb,n,flags,this);
  if (lwb>=upb)
  { push_value(std::make_shared<rational_vector_value>(RatWeight(0)));
      return; }
  const int d=arr->val.denominator();
  push_value( (flags&0x1)==0
  ? std::make_shared<rational_vector_value>(r.begin()+lwb,r.begin()+upb,d)
  : std::make_shared<rational_vector_value>(r.rbegin()+lwb,r.rbegin()+upb,d)
    );
}

template <unsigned flags>
void string_slice<flags>::evaluate(level l) const
{ int upb=(upper->eval(),get<int_value>()->val);
  int lwb=(lower->eval(),get<int_value>()->val);
  shared_string arr=(array->eval(),get<string_value>());
  const auto& r = arr->val;
  int n = r.size();
  if ((flags&0x2)!=0)
    lwb = n - lwb;
  if ((flags&0x4)!=0)
    upb = n - upb;
  if (lwb<0 or upb>n)
    slice_range_error(lwb,upb,n,flags,this);
  if (lwb>=upb)
  { push_value(std::make_shared<string_value>(std::string())); return; }
  push_value
    ((flags&0x1)==0
  ? std::make_shared<string_value>(std::string(&r[lwb],&r[upb]))
  : std::make_shared<string_value>(
       std::string(r.rbegin()+lwb,r.rbegin()+upb)));
}

template <unsigned flags>
void matrix_slice<flags>::evaluate(level l) const
{ int upb=(upper->eval(),get<int_value>()->val);
  int lwb=(lower->eval(),get<int_value>()->val);
  shared_matrix mat=(array->eval(),get<matrix_value>());
  const auto& m = mat->val;
  int n = m.numColumns();
  if ((flags&0x2)!=0)
    lwb = n - lwb;
  if ((flags&0x4)!=0)
    upb = n - upb;
  if (lwb<0 or upb>n)
    slice_range_error(lwb,upb,n,flags,this);
  if (lwb>=upb)
{ push_value(std::make_shared<matrix_value>(int_Matrix(m.numRows(),0)));
      return; }
  own_matrix result =
    std::make_shared<matrix_value>(int_Matrix(m.numRows(),upb-lwb));
  auto& r = result->val;
  for (unsigned int j=0; lwb<upb; ++j)
    r.set_column(j,m.column((flags&0x1)==0 ? lwb++ : n - ++lwb));
  push_value(std::move(result));
}

void conditional_expression::print(std::ostream& out) const
{ out << " if "; const conditional_expression* cur=this;
  do
  { out << *cur->condition << " then " << *cur->then_branch;
    conditional_expression* p =
      dynamic_cast<conditional_expression*>(cur->else_branch.get());
    if (p==nullptr)
      break;
    out << " elif "; cur=p;
  }
  while(true);
  out << " else " << *cur->else_branch << " fi ";
}

void conditional_expression::evaluate(level l) const
{ condition->eval();
  if (get<bool_value>()->val)
    then_branch->evaluate(l);
   else else_branch->evaluate(l);
}

void loop_base::print_body(std::ostream& out,unsigned flags) const
{ out << ((flags&0x1)!=0 ? " ~do " : " do ")  << *body
        << ((flags&0x2)!=0 ? " ~od " : " od ");
}

template<unsigned flags>
void while_expression<flags>::print(std::ostream& out) const
{ print_body
     (out << ((flags&0x1)!=0 ? " until " : " while ") <<  *condition
     , flags&0x2);
}

template<unsigned flags>
void while_expression<flags>::evaluate(level l) const
{ if (l==no_value)
    while (condition->eval(),get<bool_value>()->val==((flags&0x1)==0))
    { if (interrupt_flag!=0)
        throw user_interrupt();
      body->void_eval();
    }
  else
  { own_row result = std::make_shared<row_value>(0);
    auto& dst = result->val;
    while (condition->eval(),get<bool_value>()->val==((flags&0x1)==0))
    { if (interrupt_flag!=0)
        throw user_interrupt();
      body->eval();
      dst.push_back(pop_value());
    }
    if ((flags&0x2)!=0)
      std::reverse(dst.begin(),dst.end());
    push_value(std::move(result));
  }
}

template <unsigned flags, subscr_base::sub_type kind>
for_expression<flags,kind>::for_expression
 (const id_pat& p, expression_ptr&& i, expression_ptr&& b)
   : loop_base(std::move(b)), pattern(copy_id_pat(p)), in_part(i.release())
  {}

template <unsigned flags, subscr_base::sub_type kind>
void for_expression<flags,kind>::print(std::ostream& out) const
{ out << " for " << *++pattern.sublist.begin();
    if (pattern.sublist.front().kind==0x1)
      out << '@' << pattern.sublist.front();
  print_body(out << " in " << *in_part,flags);
}

template <unsigned flags, subscr_base::sub_type kind>
void for_expression<flags,kind>::evaluate(level l) const
{ in_part->eval();
  own_tuple loop_var = std::make_shared<tuple_value>(2);
       // this is safe to re-use between iterations
  own_row result;
  shared_value* dst=nullptr;
  
  switch (kind)
  {
  case subscr_base::row_entry:
    { shared_row in_val = get<row_value>();
    size_t n=in_val->val.size();
      size_t i= (flags&0x1)==0 ? 0 : n;
      if (l!=no_value)
    { result = std::make_shared<row_value>(n);
        dst = &result->val[(flags&0x2)==0 ? 0 : n];
      }
      while (i!=((flags&0x1)==0 ? n : 0))
      { loop_var->val[1]=in_val->val[(flags&0x1)==0 ? i : i-1];
          // share the current row component
        
        { loop_var->val[0] = std::make_shared<int_value>((flags&0x1)==0 ? i++ : --i);
            // index; newly created each time
          frame loop_frame (pattern);
          loop_frame.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
             *((flags&0x2)==0 ? dst++ : --dst) = pop_value();
          }
        } 
      }
    }
    break;
  case subscr_base::vector_entry:
    { shared_vector in_val = get<vector_value>();
    size_t n=in_val->val.size();
      size_t i= (flags&0x1)==0 ? 0 : n;
      if (l!=no_value)
    { result = std::make_shared<row_value>(n);
        dst = &result->val[(flags&0x2)==0 ? 0 : n];
      }
      while (i!=((flags&0x1)==0 ? n : 0))
      { loop_var->val[1] = std::make_shared<int_value>
          (in_val->val[(flags&0x1)==0 ? i : i-1]);
        
        { loop_var->val[0] = std::make_shared<int_value>((flags&0x1)==0 ? i++ : --i);
            // index; newly created each time
          frame loop_frame (pattern);
          loop_frame.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
             *((flags&0x2)==0 ? dst++ : --dst) = pop_value();
          }
        } 
      }
    }
    break;
  case subscr_base::ratvec_entry:
    { shared_rational_vector in_val = get<rational_vector_value>();
    size_t n=in_val->val.size();
      size_t i= (flags&0x1)==0 ? 0 : n;
      if (l!=no_value)
    { result = std::make_shared<row_value>(n);
        dst = &result->val[(flags&0x2)==0 ? 0 : n];
      }
      while (i!=((flags&0x1)==0 ? n : 0))
      { loop_var->val[1] = std::make_shared<rat_value> 
        (Rational
          (in_val->val.numerator()[(flags&0x1)==0 ? i : i-1]
          ,in_val->val.denominator()));
        
        { loop_var->val[0] = std::make_shared<int_value>((flags&0x1)==0 ? i++ : --i);
            // index; newly created each time
          frame loop_frame (pattern);
          loop_frame.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
             *((flags&0x2)==0 ? dst++ : --dst) = pop_value();
          }
        } 
      }
    }
    break;
  case subscr_base::string_char:
    { shared_string in_val = get<string_value>();
    size_t n=in_val->val.size();
      size_t i= (flags&0x1)==0 ? 0 : n;
      if (l!=no_value)
    { result = std::make_shared<row_value>(n);
        dst = &result->val[(flags&0x2)==0 ? 0 : n];
      }
      while (i!=((flags&0x1)==0 ? n : 0))
      { loop_var->val[1] = std::make_shared<string_value>
              (in_val->val.substr((flags&0x1)==0 ? i : i-1,1));
        
        { loop_var->val[0] = std::make_shared<int_value>((flags&0x1)==0 ? i++ : --i);
            // index; newly created each time
          frame loop_frame (pattern);
          loop_frame.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
             *((flags&0x2)==0 ? dst++ : --dst) = pop_value();
          }
        } 
      }
    }
    break;
  case subscr_base::matrix_column:
    { shared_matrix in_val = get<matrix_value>();
    size_t n=in_val->val.numColumns();
      size_t i= (flags&0x1)==0 ? 0 : n;
      if (l!=no_value)
    { result = std::make_shared<row_value>(n);
        dst = &result->val[(flags&0x2)==0 ? 0 : n];
      }
      while (i!=((flags&0x1)==0 ? n : 0))
      { loop_var->val[1] = std::make_shared<vector_value>
          (in_val->val.column((flags&0x1)==0 ? i : i-1));
        
        { loop_var->val[0] = std::make_shared<int_value>((flags&0x1)==0 ? i++ : --i);
            // index; newly created each time
          frame loop_frame (pattern);
          loop_frame.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
             *((flags&0x2)==0 ? dst++ : --dst) = pop_value();
          }
        } 
      }
    }
    break;
  case subscr_base::mod_poly_term:
    
    { shared_virtual_module pol_val = get<virtual_module_value>();
    size_t n=pol_val->val.size();
      if (l!=no_value)
    { result = std::make_shared<row_value>(n);
        dst = &result->val[(flags&0x2)==0 ? 0 : n];
      }
      if ((flags&0x1)==0)
        for (auto it=pol_val->val.cbegin(); it!=pol_val->val.cend(); ++it)
        { loop_var->val[0] =
            std::make_shared<module_parameter_value>(pol_val->rf,it->first);
          loop_var->val[1] = std::make_shared<split_int_value>(it->second);
          frame loop_frame(pattern);
          loop_frame.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
            *((flags&0x2)==0 ? dst++ : --dst) = pop_value();
          }
        } // restore context upon destruction of |loop_frame|
      else
        for (auto it=pol_val->val.crbegin(); it!=pol_val->val.crend(); ++it)
        { loop_var->val[0] =
            std::make_shared<module_parameter_value>(pol_val->rf,it->first);
          loop_var->val[1] = std::make_shared<split_int_value>(it->second);
          frame loop_frame(pattern);
          loop_frame.bind(loop_var);
          if (l==no_value)
            body->void_eval();
          else
          { body->eval();
            *((flags&0x2)==0 ? dst++ : --dst) = pop_value();
          }
        } // restore context upon destruction of |loop_frame|
    }
    break;
  case subscr_base::matrix_entry:; // excluded in type analysis
  case subscr_base::not_so: assert(false);
    }
  if (l!=no_value)
    push_value(std::move(result));
}

template <unsigned flags>
void counted_for_expression<flags>::print(std::ostream& out) const
{ print_body(out << " for " << main_hash_table->name_of(id)
                 << ": " << *count 
                 << " from " << *bound, flags);
}

template <unsigned flags>
void counted_for_expression<flags>::evaluate(level l) const
{ int b=(bound->eval(),get<int_value>()->val);
  int c=(count->eval(),get<int_value>()->val);
  if (c<0)
    c=0; // no negative size result
  if ((flags&0x1)==0)
    b+=c-1; // so that |b-c| will start at original |b|, and increase as |c--|

  if ((flags&0x4)==0)
  { id_pat pattern(id,0x1,patlist());
    if (l==no_value)
    { while (c-->0)
      { frame fr(pattern);
        fr.bind(std::make_shared<int_value>((flags&0x1)==0 ? b-c : b+c));
        body->void_eval();
      }
    }
    else
    { own_row result = std::make_shared<row_value>(c);
      auto dst = &result->val[(flags&0x2)==0 ? 0 : c];
      while (c-->0)
      { frame fr(pattern);
        fr.bind(std::make_shared<int_value>((flags&0x1)==0 ? b-c : b+c));
        body->eval();
        *((flags&0x2)==0? dst++:--dst) = pop_value();
      }
      push_value(std::move(result));
    }
  }
  else // now there is no loop counter, and no |frame| should be created
  { if (l==no_value)
    { while (c-->0)
        body->void_eval();
    }
    else
    { own_row result = std::make_shared<row_value>(c);
      auto dst = &result->val[(flags&0x2)==0 ? 0 : c];
      while (c-->0)
        { body->eval();
        *((flags&0x2)==0? dst++:--dst) = pop_value();
      }
      push_value(std::move(result));
    }
  }
}

void assignment_expr::print(std::ostream& out) const
{ out << main_hash_table->name_of(lhs) << ":=" << *rhs; }

global_assignment::global_assignment(id_type l,expression_ptr&& r)
: assignment_expr(l,std::move(r)), address(global_id_table->address_of(l)) {}

void global_assignment::evaluate(level l) const
{ rhs->eval();
  *address = pop_value();
  push_expanded(l,*address);
}

local_assignment::local_assignment
 (id_type l, size_t i,size_t j, expression_ptr&& r)
: assignment_expr(l,std::move(r)), depth(i), offset(j) {}

void local_assignment::evaluate(level l) const
{ rhs->eval();
  shared_value& dest =  frame::current->elem(depth,offset);
  dest= pop_value();
  push_expanded(l,dest);
}

template <bool reversed>
void component_assignment<reversed>::print(std::ostream& out) const
{ out << main_hash_table->name_of(lhs) << (reversed ? "~[" : "[")
        << *index << "]:=" << *rhs;
}

template <bool reversed>
global_component_assignment<reversed>::global_component_assignment 
  (id_type a,expression_ptr&& i,expression_ptr&& r, subscr_base::sub_type k)
: base(a,std::move(i),std::move(r))
, kind(k),address(global_id_table->address_of(a)) {}

template <bool reversed>
void global_component_assignment<reversed>::evaluate(expression_base::level l)
  const
{ if (address->get()==nullptr)
  { std::ostringstream o;
    o << "Assigning to component of uninitialized variable "
      << main_hash_table->name_of(this->lhs);
    throw runtime_error(o.str());
  }
  base::assign(l,*address,kind);
}

template <bool reversed>
void component_assignment<reversed>::assign
  (level lev,shared_value& aggregate, subscr_base::sub_type kind) const
{ rhs->eval();
  value loc=uniquify(aggregate);
    // simple pointer to modifiable value from shared pointer
  switch (kind)
  { case subscr_base::row_entry:
  
  { unsigned int i=(index->eval(),get<int_value>()->val);
    std::vector<shared_value>& a=force<row_value>(loc)->val;
    size_t n=a.size();
    if (i>=n)
      throw runtime_error(range_mess(i,a.size(),this));
    auto& ai = a[reversed ? n-1-i : i];
    ai = pop_value(); // assign non-expanded value
    push_expanded(lev,ai); // return value may need expansion, or be omitted
  }
  break;
    case subscr_base::vector_entry:
  
  { unsigned int i=(index->eval(),get<int_value>()->val);
    std::vector<int>& v=force<vector_value>(loc)->val;
    size_t n=v.size();
    if (i>=n)
      throw runtime_error(range_mess(i,v.size(),this));
    v[reversed ? n-1-i : i]= force<int_value>(execution_stack.back().get())->val;
      // assign |int| from un-popped top
    if (lev==no_value)
      execution_stack.pop_back(); // pop it anyway if result not needed
  }
  break;
    case subscr_base::matrix_entry:
  
  { index->multi_eval();
    unsigned int j=get<int_value>()->val;
    unsigned int i=get<int_value>()->val;
  
    int_Matrix& m=force<matrix_value>(loc)->val;
    size_t k=m.numRows(),l=m.numColumns();
    if (i>=k)
      throw runtime_error(range_mess(i,m.numRows(),this));
    if (j>=l)
      throw runtime_error(range_mess(j,m.numColumns(),this));
    m(reversed ? k-1-i : i,reversed ? l-1-j : j)=
      force<int_value>(execution_stack.back().get())->val;
      // assign |int| from un-popped top
    if (lev==no_value)
      execution_stack.pop_back(); // pop it anyway if result not needed
  }
  break;
    case subscr_base::matrix_column:
  
  { unsigned int j=(index->eval(),get<int_value>()->val);
    int_Matrix& m=force<matrix_value>(loc)->val;
  const int_Vector& v=force<vector_value>(execution_stack.back().get())->val;
      // don't pop
    size_t l=m.numColumns();
    if (j>=l)
      throw runtime_error(range_mess(j,m.numColumns(),this));
    if (v.size()!=m.numRows())
      throw runtime_error
        (std::string("Cannot replace column of size ")+str(m.numRows())+
         " by one of size "+str(v.size()));
    m.set_column(reversed ? l-j-1 : j,v);
      // copy value of |int_Vector| into the matrix
    if (lev==no_value)
      execution_stack.pop_back(); // pop the vector if result not needed
  }
  break;
  default: {} // remaining cases are eliminated in type analysis
  }
}

template <bool reversed>
local_component_assignment<reversed>::local_component_assignment
 (id_type arr, expression_ptr&& i,size_t d, size_t o, expression_ptr&& r,
  subscr_base::sub_type k)
: base(arr,std::move(i),std::move(r)), kind(k), depth(d), offset(o) {}

template <bool reversed>
void local_component_assignment<reversed>::evaluate(expression_base::level l)
  const
{ base::assign (l,frame::current->elem(depth,offset),kind); }

} }

